header {
package scheme.parser;
import scheme.absyn.*;
import scheme.util.*;
}

//------------------------------------------------------------------------------------------------------
// Lexical analysis
//------------------------------------------------------------------------------------------------------

class SchemeLexer extends Lexer;

options {
        exportVocab=Scheme;
        k=2;
}

TRUE       : "#t"               ;
FALSE      : "#f"               ;
LEQ        : "<="               ;
QUOTECHAR  : '\''               ;
LPAREN     : '('|'['            ;
RPAREN     : ')'|']'            ;
MINUS      : '-'                ;
DOT        : '.'                ;
TIMES      : '*'                ;
PLUS       : '+'                ;
LESS       : '<'                ;
EQ         : '='                ;

ID	   :	('a'..'z'|'A'..'Z'|'_'|'$') ('a'..'z'|'A'..'Z'|'_'|'-'|'!'|'?'|'0'..'9'|'$')*
	   ;

INT        :    ('0'..'9')+ 
           ;

// Whitespace
WS	   :	(	' '
		|	'\t'
		|	'\f'
		// handle newlines
		|	(	"\r\n"  // Evil DOS
			|	'\r'    // Macintosh
			|	'\n'    // Unix (the right way)
			)
			{ newline(); }
		)
		{ $setType(Token.SKIP); }
	   ;

// Comments
COMMENT    :	";"
		(~('\n'|'\r'))* ('\n'|'\r'('\n')?)
		{$setType(Token.SKIP); newline();}
	   ;

//------------------------------------------------------------------------------------------------------
// Parser...
//------------------------------------------------------------------------------------------------------

class SchemeParser extends Parser;

options {
        exportVocab=Scheme;
        k=2;
        buildAST=true;
}

tokens { 
       APP; PRIM; EMPTY_LIST; DECL; PARAMS; 
} 

exp 
 : 
   EOF!
 | TRUE  
 | FALSE
 | INT
 | ID
 | LPAREN! "quote"^ datum RPAREN!
 | QUOTECHAR^ datum 
 | LPAREN! "set!"^ ID exp RPAREN!
 | LPAREN! "if"^ exp exp exp RPAREN!
 | LPAREN! "begin"^ (exp)* RPAREN!
 | LPAREN! "let"^ LPAREN! (decl)* RPAREN! exp RPAREN!
 | LPAREN! "letrec"^ LPAREN! (decl)* RPAREN! exp RPAREN!
 | LPAREN! "lambda"^ params exp RPAREN! 
 | plp:LPAREN^ prim (exp)* RPAREN! { #plp.setType(PRIM); }
 | alp:LPAREN^ exp (exp)* RPAREN! { #alp.setType(APP); }
 ;

params
 : 
   lp:LPAREN^ (ID)* RPAREN! { #lp.setType(PARAMS); }
 ; 

datum
 : 
   TRUE
 | FALSE
 | INT
 | lp:LPAREN^ RPAREN! { #lp.setType(EMPTY_LIST); }
 ;

decl
 : 
   lp:LPAREN^ ID exp RPAREN! { #lp.setType(DECL); }
 ;

prim
 : 
   MINUS
 | TIMES
 | PLUS
 | LESS
 | LEQ
 | EQ
 | "add1"
 | "boolean?"
 | "car"
 | "cdr"
 | "cons"
 | "integer?"
 | "not"
 | "null?"
 | "pair?"
 | "procedure?"
 | "sub1"
 | "void"
 | "zero?"
 | "eq?"
 | "set-car!"
 | "set-cdr!"
 ; 

//------------------------------------------------------------------------------------------------------
// Transform the input tree to an AST...
//------------------------------------------------------------------------------------------------------

class SchemeTreeParser extends TreeParser;

exp returns [Exp e] {
     e = null;
     byte primop;
     ExpList exps = new EmptyExpL();
     List decls = new EmptyL();
     Exp se, e1, e2, e3, be, letb, letrecb, lamb, pa, fun, arg;
     Const c;
     Datum d, qd;
     Decl letd, letrecd;
     List ps = new EmptyL();
}
 : 
   c=constant { e = new ConstantE (c); }
 | i:ID { e = new VarE(i.getText()); }
 | #("quote" d=datum) { e = new QuoteE (d); }
 | #(QUOTECHAR qd=datum) { e = new QuoteE (qd); }
 | #("set!" si:ID se=exp) { e = new SetE(si.getText(),se); }
 | #("if" e1=exp e2=exp e3=exp) { e = new IfE(e1,e2,e3); }
 | #("begin" (be=exp { exps = exps.snoc(be); })*) { e = new BeginE (exps); }
 | #("let" (letd=decl { decls = decls.snoc(letd); })* letb=exp) { e = new LetE(decls,letb); }
 | #("letrec" (letrecd=decl { decls = decls.snoc(letrecd); })* letrecb=exp) 
     { e = new LetrecE(decls,letrecb); }
 | #("lambda" ps=params lamb=exp) { e = new LambdaE(ps,lamb); }
 | #(PRIM primop=prim (pa=exp { exps = exps.snoc(pa); })*) { e = new PrimappE (primop,exps); }
 | #(APP fun=exp (arg=exp { exps = exps.snoc(arg); })*) { e = new AppE (fun,exps); }
 ;

params returns [List ids] {
       ids = new EmptyL();
}
 : 
   #(PARAMS (id:ID { ids = ids.snoc(id.getText()); })*) 
 ;

constant returns [Const c] {
         c = null;
}
 : 
   TRUE { c = new BoolC(true); }
 | FALSE { c = new BoolC(false); }
 | i:INT { c = new IntC(Integer.parseInt(i.getText())); }
 ;

datum returns [Datum datum] {
      datum = null;
      Const c;
}
 : 
   c=constant { datum = new ConstantD (c); }
 | EMPTY_LIST { datum = new EmptyListD(); }
 ;

decl returns [Decl decl] {
     decl = null;
     Exp rhs;
}    
 : 
   #(DECL i:ID rhs=exp) { decl = new Decl(i.getText(),rhs); }
 ;

prim returns [byte primb] {
     primb = (byte)(-1);
}
 : 
   MINUS { primb = Prim.MINUS; }
 | TIMES { primb = Prim.TIMES; }
 | PLUS { primb = Prim.PLUS; }
 | LESS { primb = Prim.LESS; }
 | LEQ { primb = Prim.LEQ; }
 | EQ { primb = Prim.EQ; }
 | "add1" { primb = Prim.ADD1; }
 | "boolean?" { primb = Prim.BOOLEANP; }
 | "car" { primb = Prim.CAR; }
 | "cdr" { primb = Prim.CDR; }
 | "cons" { primb = Prim.CONS; }
 | "integer?" { primb = Prim.INTEGERP; }
 | "not" { primb = Prim.NOT; }
 | "null?" { primb = Prim.NULLP; }
 | "pair?" { primb = Prim.PAIRP; }
 | "procedure?" { primb = Prim.PROCEDUREP; }
 | "sub1" { primb = Prim.SUB1; }
 | "void" { primb = Prim.VOID; }
 | "zero?" { primb = Prim.ZEROP; }
 | "eq?" { primb = Prim.EQP; }
 | "set-car!" { primb = Prim.SETCAR; }
 | "set-cdr!" { primb = Prim.SETCDR; }
 ; 

